/*
 *  Copyright (c) 2002 Petko Manolov (petkan@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * version 2 as published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/mii.h>
#include <linux/ethtool.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/usb.h>
#include <linux/init.h>
#include <linux/cache.h>
#include <asm/uaccess.h>

#ifdef DEBUG
#undef dbg
#define dbg(format, arg...) printk(KERN_INFO __FILE__ "!"  __FUNCTION__ ": " format "\n" , ## arg)
static int msglvl;
#endif

/* Version Information */
#define DRIVER_VERSION "v0.4.1 (2002/07/22)"
#define DRIVER_AUTHOR "Petko Manolov <petkan@users.sourceforge.net>"
#define DRIVER_DESC "rtl8150 based usb-ethernet driver"

#define	IRD			0x0120
#define	MAR			0x0126
#define	CR			0x012e
#define	TCR			0x012f
#define	RCR			0x0130
#define	TSR			0x0132
#define	RSR			0x0133
#define	CON0			0x0135
#define	CON1			0x0136
#define	MSR			0x0137
#define	PHYADD			0x0138
#define	PHYDAT			0x0139
#define	PHYCNT			0x013b
#define	GPPC			0x013d
#define	BMCR			0x0140
#define	BMSR			0x0142
#define	ANAR			0x0144
#define	ANLP			0x0146
#define	AER			0x0148

#define FIRST_REG		0x0120
#define LAST_REG		0x018A

#define RTL8150_REGS_SIZE	(LAST_REG - FIRST_REG)
#define RTL8150_REGS_VER	(1)

#define RTL8150_EEPROM_BASE	(0x1200)
#define RTL8150_EEPROM_SIZE	(0x80)

#define	PHY_READ		0
#define	PHY_WRITE		0x20
#define	PHY_GO			0x40

#define	RTL8150_REQT_READ	0xc0
#define	RTL8150_REQT_WRITE	0x40
#define	RTL8150_REQ_GET_REGS	0x05
#define	RTL8150_REQ_SET_REGS	0x05

#define	RTL8150_MTU		1500
#define	RTL8150_MAX_MTU		1536

#define	RTL8150_TX_TIMEOUT	(HZ)


/* rtl8150 flags */
#define	RTL8150_FLAG_HWCRC	0
#define	RX_REG_SET		1
#define	RTL8150_UNPLUG		2

/* Define these values to match your device */
#define VENDOR_ID_REALTEK		0x0bda
#define	VENDOR_ID_MELCO			0x0411

#define PRODUCT_ID_RTL8150		0x8150
#define	PRODUCT_ID_LUAKTX		0x0012

/* table of devices that work with this driver */
static struct usb_device_id rtl8150_table[] = {
	{USB_DEVICE(VENDOR_ID_REALTEK, PRODUCT_ID_RTL8150)},
	{USB_DEVICE(VENDOR_ID_MELCO, PRODUCT_ID_LUAKTX)},
	{}
};

MODULE_DEVICE_TABLE(usb, rtl8150_table);

struct rtl8150 {
	unsigned int flags;
	struct usb_device *udev;
	struct usb_interface *interface;
	struct semaphore sem;
	struct net_device_stats stats;
	struct net_device *netdev;
	struct urb *rx_urb, *tx_urb, *intr_urb, *ctrl_urb;
	int intr_interval;
	unsigned long poll_cnt;
	unsigned long link_chg;
	unsigned long link_down;
	unsigned long link_up;
	u8 link_sts;
	u8 phy;
	struct usb_ctrlrequest dr ____cacheline_aligned;
	u16 rx_creg;	/* this can share cache line with dr */
	u8 rx_buff[RTL8150_MAX_MTU] ____cacheline_aligned;
	u8 tx_buff[RTL8150_MAX_MTU] ____cacheline_aligned;
	u8 intr_buff[8] ____cacheline_aligned;
};

typedef struct rtl8150 rtl8150_t;

/* the global usb devfs handle */
extern devfs_handle_t usb_devfs_handle;
/*
 * Since this driver doesn't currently support multicast filtering,
 * set ALL_MULTI if any multicast addresses specified
 */
unsigned long multicast_filter_limit = 0;

static void rtl8150_disconnect(struct usb_device *dev, void *ptr);
static void *rtl8150_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id);

static struct usb_driver rtl8150_driver = {
	name:		"rtl8150",
	probe:		rtl8150_probe,
	disconnect:	rtl8150_disconnect,
	id_table:	rtl8150_table,
};

/*
**
**	device related part of the code
**
*/
static int get_registers(rtl8150_t * dev, u16 indx, u16 size, void *data)
{
	return usb_control_msg(dev->udev, usb_rcvctrlpipe(dev->udev, 0),
			       RTL8150_REQ_GET_REGS, RTL8150_REQT_READ,
			       indx, 0, data, size, HZ / 2);
}

static int set_registers(rtl8150_t * dev, u16 indx, u16 size, void *data)
{
	return usb_control_msg(dev->udev, usb_sndctrlpipe(dev->udev, 0),
			       RTL8150_REQ_SET_REGS, RTL8150_REQT_WRITE,
			       indx, 0, data, size, HZ / 2);
}

static void ctrl_callback(struct urb *urb)
{
	rtl8150_t *dev;

	switch (urb->status) {
	case 0:
		break;
	case -EINPROGRESS:
		break;
	case -ENOENT:
		break;
	default:
		warn("ctrl urb status %d", urb->status);
	}
	dev = urb->context;
	clear_bit(RX_REG_SET, &dev->flags);
}

static int async_set_registers(rtl8150_t * dev, u16 indx, u16 size, void *data)
{
	int ret;

	if (test_bit(RX_REG_SET, &dev->flags))
		return -EAGAIN;

	dev->dr.bRequestType = RTL8150_REQT_WRITE;
	dev->dr.bRequest = RTL8150_REQ_SET_REGS;
	dev->dr.wValue = cpu_to_le16(indx);
	dev->dr.wIndex = 0;
	dev->dr.wLength = cpu_to_le16(2);
	dev->ctrl_urb->transfer_buffer_length = 2;
	FILL_CONTROL_URB(dev->ctrl_urb, dev->udev,
			 usb_sndctrlpipe(dev->udev, 0), (char *) &dev->dr,
			 &dev->rx_creg, 2, ctrl_callback, dev);
	if ((ret = usb_submit_urb(dev->ctrl_urb)))
		err("control request submission failed: %d", ret);
	else
		set_bit(RX_REG_SET, &dev->flags);

	return ret;
}

static int read_mii_word(rtl8150_t * dev, u8 phy, __u8 indx, u16 * reg)
{
	int i;
	u8 data[3], tmp;

	data[0] = phy;
	data[1] = data[2] = 0;
	tmp = indx | PHY_READ | PHY_GO;
	i = 0;

	set_registers(dev, PHYADD, sizeof(data), data);
	set_registers(dev, PHYCNT, 1, &tmp);
	do {
		get_registers(dev, PHYCNT, 1, data);
	} while ((data[0] & PHY_GO) && (i++ < HZ));

	if (i < HZ) {
		get_registers(dev, PHYDAT, 2, data);
		*reg = le16_to_cpup((u16 *)data);
		return 0;
	} else
		return 1;
}

static int write_mii_word(rtl8150_t * dev, u8 phy, __u8 indx, u16 reg)
{
	int i;
	u8 data[3], tmp;

	data[0] = phy;
	*(data + 1) = cpu_to_le16p(&reg);
	tmp = indx | PHY_WRITE | PHY_GO;
	i = 0;

	set_registers(dev, PHYADD, sizeof(data), data);
	set_registers(dev, PHYCNT, 1, &tmp);
	do {
		get_registers(dev, PHYCNT, 1, data);
	} while ((data[0] & PHY_GO) && (i++ < HZ));

	if (i < HZ)
		return 0;
	else
		return 1;
}

static inline void set_ethernet_addr(rtl8150_t * dev)
{
	u8 node_id[6];

	get_registers(dev, IRD, sizeof(node_id), node_id);
	memcpy(dev->netdev->dev_addr, node_id, sizeof(node_id));
}

static int rtl8150_reset(rtl8150_t * dev)
{
	u8 data = 0x10;
	int i = HZ;

	set_registers(dev, CR, 1, &data);
	do {
		get_registers(dev, CR, 1, &data);
	} while ((data & 0x10) && --i);

	return (i > 0) ? 0 : -1;
}

static int alloc_all_urbs(rtl8150_t * dev)
{
	dev->rx_urb = usb_alloc_urb(0);
	if (!dev->rx_urb)
		return 0;
	dev->tx_urb = usb_alloc_urb(0);
	if (!dev->tx_urb) {
		usb_free_urb(dev->rx_urb);
		return 0;
	}
	dev->intr_urb = usb_alloc_urb(0);
	if (!dev->intr_urb) {
		usb_free_urb(dev->rx_urb);
		usb_free_urb(dev->tx_urb);
		return 0;
	}
	dev->ctrl_urb = usb_alloc_urb(0);
	if (!dev->ctrl_urb) {
		usb_free_urb(dev->rx_urb);
		usb_free_urb(dev->tx_urb);
		usb_free_urb(dev->intr_urb);
		return 0;
	}

	return 1;
}

static void free_all_urbs(rtl8150_t * dev)
{
	usb_free_urb(dev->rx_urb);
	usb_free_urb(dev->tx_urb);
	usb_free_urb(dev->intr_urb);
	usb_free_urb(dev->ctrl_urb);
}

static void unlink_all_urbs(rtl8150_t * dev)
{
	usb_unlink_urb(dev->rx_urb);
	usb_unlink_urb(dev->tx_urb);
	usb_unlink_urb(dev->intr_urb);
	usb_unlink_urb(dev->ctrl_urb);
}

static void read_bulk_callback(struct urb *urb)
{
	rtl8150_t *dev;
	int pkt_len, res;
	struct sk_buff *skb;
	struct net_device *netdev;
	u16 rx_stat;

	dev = urb->context;
	if (!dev) {
		warn("!dev");
		return;
	}
	netdev = dev->netdev;
	if (!netif_device_present(netdev)) {
		warn("netdev is not present");
		return;
	}
	switch (urb->status) {
	case 0:
		break;
	case -ENOENT:
		return;
	case -ETIMEDOUT:
		warn("Rx timeout, reset needed may be?..");
		return;
	default:
		warn("Rx status %d", urb->status);
		goto goon;
	}

	pkt_len = urb->actual_length - 4;
	rx_stat = le16_to_cpu(*(u16 *) (dev->rx_buff + pkt_len));

	if (!(skb = dev_alloc_skb(pkt_len + 2)))
		goto goon;
	skb->dev = netdev;
	skb_reserve(skb, 2);
	eth_copy_and_sum(skb, dev->rx_buff, pkt_len, 0);
	skb_put(skb, pkt_len);
	skb->protocol = eth_type_trans(skb, netdev);
	netif_rx(skb);
	dev->stats.rx_packets++;
	dev->stats.rx_bytes += pkt_len;
goon:
	FILL_BULK_URB(dev->rx_urb, dev->udev, usb_rcvbulkpipe(dev->udev, 1),
		      dev->rx_buff, RTL8150_MAX_MTU, read_bulk_callback, dev);
	if ((res = usb_submit_urb(dev->rx_urb)))
		warn("%s: Rx urb submission failed %d", netdev->name, res);
}

static void write_bulk_callback(struct urb *urb)
{
	rtl8150_t *dev;

	dev = urb->context;
	if (!dev)
		return;
	if (!netif_device_present(dev->netdev))
		return;
	if (urb->status)
		info("%s: Tx status %d", dev->netdev->name, urb->status);
	dev->netdev->trans_start = jiffies;
	netif_wake_queue(dev->netdev);
}

void intr_callback(struct urb *urb)
{
	rtl8150_t *dev;
	__u8 *d;

	dev = urb->context;
	if (!dev)
		return;
	switch (urb->status) {
	case 0:
		break;
	case -ENOENT:
		return;
	default:
		info("%s: intr status %d", dev->netdev->name, urb->status);
	}

	d = urb->transfer_buffer;

	if ( (d[0] & 0x3c) || d[7] ) {
		dev->stats.tx_errors++;

		if ( d[0] & 0x20 )
			dev->stats.collisions++;
		if ( d[0] & 0x10 )
			dev->stats.tx_window_errors++;
		if ( d[0] & 0x08 )
			dev->stats.tx_carrier_errors++;
		if ( d[0] & 0x04 )
			dev->stats.tx_aborted_errors++;

		dev->stats.collisions += d[7];
	}

	if ( (d[1] & 0x1E) || d[5] || d[6] ) {
		dev->stats.rx_errors++;

		if ( d[1] & 0x10 )
			dev->stats.rx_length_errors++;
		if ( d[1] & 0x08 )
			dev->stats.rx_length_errors++;
		if ( d[1] & 0x04 )
			dev->stats.rx_crc_errors++;
		if ( d[1] & 0x02 )
			dev->stats.rx_frame_errors++;

		dev->stats.rx_fifo_errors += d[5];
		dev->stats.rx_crc_errors += d[6];
	}

	if ( d[1] * 0x20 )
		dev->link_chg++;

	if ( (d[2] & 0x4) ) {		// link is up
		if ( !dev->link_sts ) {	// was down
		    dev->link_up++;
		    dev->link_sts = 1;
		    netif_carrier_on(dev->netdev);
		    info("link up 10%s%sDX", (d[2]&0x8)?"0":"", (d[2]&0x10)?"F":"H");
		}
	}
	else {				// link is down
		if ( dev->link_sts ) {	// was up
		    dev->link_down++;
		    dev->link_sts = 0;
		    netif_carrier_off(dev->netdev);
		    info("link down");
		}
	}
	dev->poll_cnt++;
}

/*
**
**	network related part of the code
**
*/

static int enable_net_traffic(rtl8150_t * dev)
{
	u8 cr, tcr, rcr, msr;

	if (rtl8150_reset(dev)) {
		warn("%s - device reset failed", __FUNCTION__);
	}
	dev->rx_creg = cpu_to_le16(0x009e);
	rcr = 0x9e;	/* bit7=1 attach Rx info at the end */
	tcr = 0xd8;
	cr = 0x0e;
	set_registers(dev, RCR, 1, &rcr);
	set_registers(dev, TCR, 1, &tcr);
	set_registers(dev, CR, 1, &cr);
	get_registers(dev, MSR, 1, &msr);

	return 0;
}

static void disable_net_traffic(rtl8150_t * dev)
{
	u8 cr;

	get_registers(dev, CR, 1, &cr);
	cr &= 0xf3;
	set_registers(dev, CR, 1, &cr);
}

static struct net_device_stats *rtl8150_netdev_stats(struct net_device *dev)
{
#ifdef DEBUG
	rtl8150_t *r = (rtl8150_t *) dev->priv;

	dbg("link=%d up=%lu dn=%lu chg=%lu poll_cnt=%lu",
	    r->link_sts, r->link_up, r->link_down, r->link_chg, r->poll_cnt);
#endif
	return &((rtl8150_t *) dev->priv)->stats;
}

static void rtl8150_tx_timeout(struct net_device *netdev)
{
	rtl8150_t *dev;

	dev = netdev->priv;
	if (!dev)
		return;
	warn("%s: Tx timeout.", netdev->name);
	dev->tx_urb->transfer_flags |= USB_ASYNC_UNLINK;
	usb_unlink_urb(dev->tx_urb);
	dev->stats.tx_errors++;
}

static void rtl8150_set_multicast(struct net_device *netdev)
{
	rtl8150_t *dev;

	dev = netdev->priv;
	if (netdev->flags & IFF_PROMISC) {
		dev->rx_creg |= cpu_to_le16(0x0001);
		info("%s: promiscuous mode", netdev->name);
	} else if ((netdev->mc_count > multicast_filter_limit) ||
		   (netdev->flags & IFF_ALLMULTI)) {
		dev->rx_creg &= cpu_to_le16(0xfffe);
		dev->rx_creg |= cpu_to_le16(0x0002);
		info("%s: allmulti set", netdev->name);
	} else {
		/* ~RX_MULTICAST, ~RX_PROMISCUOUS */
		dev->rx_creg &= cpu_to_le16(0x00fc);
	}
	async_set_registers(dev, RCR, 2, &dev->rx_creg);
}

static int rtl8150_start_xmit(struct sk_buff *skb, struct net_device *netdev)
{
	rtl8150_t *dev;
	int count, res;

	netif_stop_queue(netdev);
	dev = netdev->priv;
	count = (skb->len < 60) ? 60 : skb->len;
	if (count & 0x3f) {
		dev->tx_urb->transfer_flags &= ~USB_ZERO_PACKET;
	} else {
		dev->tx_urb->transfer_flags |= USB_ZERO_PACKET;
	}

	memcpy(dev->tx_buff, skb->data, skb->len);
	if (skb->len < count) {
		memset(dev->tx_buff + skb->len, 0, count - skb->len);
	}
	FILL_BULK_URB(dev->tx_urb, dev->udev, usb_sndbulkpipe(dev->udev, 2),
		      dev->tx_buff, RTL8150_MAX_MTU, write_bulk_callback, dev);
	dev->tx_urb->transfer_buffer_length = count;

	if ((res = usb_submit_urb(dev->tx_urb))) {
		warn("failed tx_urb %d\n", res);
		dev->stats.tx_errors++;
		netif_start_queue(netdev);
	} else {
		dev->stats.tx_packets++;
		dev->stats.tx_bytes += skb->len;
		netdev->trans_start = jiffies;
	}
	dev_kfree_skb(skb);

	return 0;
}

static int rtl8150_open(struct net_device *netdev)
{
	rtl8150_t *dev;
	int res;

	dbg("<<");
	dev = netdev->priv;
	if (dev == NULL) {
		return -ENODEV;
	}

	down(&dev->sem);
	FILL_BULK_URB(dev->rx_urb, dev->udev, usb_rcvbulkpipe(dev->udev, 1),
		      dev->rx_buff, RTL8150_MAX_MTU, read_bulk_callback, dev);
	if ((res = usb_submit_urb(dev->rx_urb)))
		warn("%s: rx_urb submit failed: %d", __FUNCTION__, res);
	FILL_INT_URB(dev->intr_urb, dev->udev, usb_rcvintpipe(dev->udev, 3),
		     dev->intr_buff, sizeof(dev->intr_buff), intr_callback,
		     dev, dev->intr_interval);
	if ((res = usb_submit_urb(dev->intr_urb)))
		warn("%s: intr_urb submit failed: %d", __FUNCTION__, res);
	netif_start_queue(netdev);
	enable_net_traffic(dev);
	up(&dev->sem);

	dbg(">>");
	return res;
}

static int rtl8150_close(struct net_device *netdev)
{
	rtl8150_t *dev;

	dbg("<<");
	dev = netdev->priv;
	if (!dev)
		return -ENODEV;

	down(&dev->sem);
		disable_net_traffic(dev);
	netif_stop_queue(netdev);

	unlink_all_urbs(dev);
	up(&dev->sem);
	dbg(">>");

	return 0;
}

static int rtl8150_ethtool_ioctl(struct net_device *netdev, void *uaddr)
{
	rtl8150_t *dev;
	int cmd;
	char tmp[128];

	dev = netdev->priv;
	if (get_user(cmd, (int *) uaddr))
		return -EFAULT;

	switch (cmd) {
	case ETHTOOL_GDRVINFO:{
		struct ethtool_drvinfo info = { ETHTOOL_GDRVINFO };

		strncpy(info.driver, DRIVER_DESC, ETHTOOL_BUSINFO_LEN);
		strncpy(info.version, DRIVER_VERSION,
			ETHTOOL_BUSINFO_LEN);
		sprintf(tmp, "usb%d:%d", dev->udev->bus->busnum,
			dev->udev->devnum);
		strncpy(info.bus_info, tmp, ETHTOOL_BUSINFO_LEN);
		info.eedump_len = RTL8150_EEPROM_SIZE;
		info.regdump_len = RTL8150_REGS_SIZE;
		if (copy_to_user(uaddr, &info, sizeof(info)))
			return -EFAULT;
		return 0;
		}
	case ETHTOOL_GSET:{
		struct ethtool_cmd ecmd;
		short lpa, bmcr, ana;

		if (copy_from_user(&ecmd, uaddr, sizeof(ecmd)))
			return -EFAULT;
		ecmd.supported = (SUPPORTED_10baseT_Half |
				  SUPPORTED_10baseT_Full |
				  SUPPORTED_100baseT_Half |
				  SUPPORTED_100baseT_Full |
				  SUPPORTED_Autoneg |
				  SUPPORTED_TP | SUPPORTED_MII);
		ecmd.port = PORT_TP;
		ecmd.transceiver = XCVR_INTERNAL;
		ecmd.phy_address = dev->phy;
		get_registers(dev, BMCR, 2, &bmcr);
		bmcr = le16_to_cpu(bmcr);
		if (bmcr & BMCR_ANENABLE) {
			get_registers(dev, ANLP, 2, &lpa);
			lpa = le16_to_cpu(lpa);
			get_registers(dev, ANAR, 2, &ana);
			ana = le16_to_cpu(ana);
			ecmd.advertising = ADVERTISED_Autoneg;
			ecmd.advertising |= (ana | ADVERTISE_10HALF) ? ADVERTISED_10baseT_Half : 0;
			ecmd.advertising |= (ana | ADVERTISE_10FULL) ? ADVERTISED_10baseT_Full : 0;
			ecmd.advertising |= (ana | ADVERTISE_100HALF) ? ADVERTISED_100baseT_Half : 0;
			ecmd.advertising |= (ana | ADVERTISE_100FULL) ? ADVERTISED_100baseT_Full : 0;
			ecmd.autoneg = AUTONEG_ENABLE;
			ecmd.speed =
				(lpa & (LPA_100HALF | LPA_100FULL)) ?
				SPEED_100 : SPEED_10;
			if (ecmd.speed == SPEED_100)
				ecmd.duplex = (lpa & LPA_100FULL) ?
				    DUPLEX_FULL : DUPLEX_HALF;
			else
				ecmd.duplex = (lpa & LPA_10FULL) ?
				    DUPLEX_FULL : DUPLEX_HALF;
		} else {
			ecmd.autoneg = AUTONEG_DISABLE;
			ecmd.speed = (bmcr & BMCR_SPEED100) ?
				SPEED_100 : SPEED_10;
			ecmd.duplex = (bmcr & BMCR_FULLDPLX) ?
				DUPLEX_FULL : DUPLEX_HALF;
		}
		if (copy_to_user(uaddr, &ecmd, sizeof(ecmd)))
			return -EFAULT;
		return 0;
		}
	case ETHTOOL_SSET:
		return -ENOTSUPP;
	case ETHTOOL_GLINK:{
		struct ethtool_value edata = { ETHTOOL_GLINK };

		edata.data = netif_carrier_ok(netdev);
		if (copy_to_user(uaddr, &edata, sizeof(edata)))
			return -EFAULT;
		return 0;
		}
	/* get registers */
	case ETHTOOL_GREGS: {
		struct ethtool_regs regs;
		u8 regbuf[RTL8150_REGS_SIZE];
		u8 d[4];
		int r, i, x;

		if (copy_from_user(&regs, uaddr, sizeof(regs)))
			return -EFAULT;

		if (regs.len > RTL8150_REGS_SIZE) {
			regs.len = RTL8150_REGS_SIZE;
		}
		dbg("regs.len = %x\n", regs.len);

		regs.version = RTL8150_REGS_VER;
		if (copy_to_user(uaddr, &regs, sizeof(regs)))
			return -EFAULT;


		x = FIRST_REG;
		i = 0;
		while ( i < regs.len  ) {
			r = get_registers(dev, x, 2, d);
			dbg("i=%x x=%x d0=%x d1=%x r=%d", 
				i, x, d[0], d[1], r);
			regbuf[i] = d[0];
			regbuf[i+1] = d[1];
			i += 2;
			x += 2;
		}

		uaddr += offsetof(struct ethtool_regs, data);
		if (copy_to_user(uaddr, regbuf, regs.len))
			return -EFAULT;
		return 0;
        }
#ifdef DEBUG
	/* get eprom */
	case ETHTOOL_GEEPROM: {
		struct ethtool_eeprom eeprom;
		u8 eprombuf[RTL8150_EEPROM_SIZE];
		u8 d[4];
		int r, i, x;

		if (copy_from_user(&eeprom, uaddr, sizeof(eeprom)))
			return -EFAULT;

		if (eeprom.offset + eeprom.len > RTL8150_EEPROM_SIZE) {
			return -EFAULT;
		}

		x = eeprom.offset + RTL8150_EEPROM_BASE;
		i = 0;
		while ( x < RTL8150_EEPROM_SIZE  ) {
			r = get_registers(dev, x, 2, d);
			dbg("eprom x=%x d0=%x d1=%x r=%d", 
				x, d[0], d[1], r);
			eprombuf[i] = d[0];
			eprombuf[i+1] = d[1];
			i += 2;
			x += 2;
		}
		uaddr += offsetof(struct ethtool_eeprom, data);
		if (copy_to_user(uaddr, eprombuf, eeprom.len))
			return -EFAULT;
		return 0;
	}

	/* get message-level */
	case ETHTOOL_GMSGLVL: {
		struct ethtool_value edata = {ETHTOOL_GMSGLVL};
		edata.data = msglvl;
		dbg("link=%d up=%lu dn=%lu poll_cnt=%lu",
		    dev->link_sts, dev->link_up, dev->link_down, dev->poll_cnt);
		if (copy_to_user(uaddr, &edata, sizeof(edata)))
			return -EFAULT;
		return 0;
	}
#endif
	/* set message-level */
	case ETHTOOL_SMSGLVL: {
		struct ethtool_value edata;
		void usb_hub_port_reconnect(struct usb_device *);

		if (copy_from_user(&edata, uaddr, sizeof(edata)))
			return -EFAULT;
		if (edata.data == 99)
			usb_hub_port_reconnect(dev->udev);
#ifdef DEBUG
		else
			msglvl = edata.data; 
#endif
		return 0;
	}
	/* restart autonegotiation */
	case ETHTOOL_NWAY_RST: {
		__u16 bmcr;
		info("restarting auto-negotiations");
		get_registers(dev, BMCR, 2, &bmcr);
		bmcr = le16_to_cpu(bmcr);
		bmcr |= BMCR_ANRESTART;
		bmcr = cpu_to_le16(bmcr);
		set_registers(dev, BMCR, 2, &bmcr);
		return 0;
	}
	default:
		return -EOPNOTSUPP;
	}
}

static int rtl8150_ioctl(struct net_device *netdev, struct ifreq *rq, int cmd)
{
	rtl8150_t *dev;
	u16 *data;
	int res;

	dev = netdev->priv;
	data = (u16 *) & rq->ifr_data;
	res = 0;

	down(&dev->sem);
	switch (cmd) {
	case SIOCETHTOOL:
		res = rtl8150_ethtool_ioctl(netdev, rq->ifr_data);
		break;
	case SIOCDEVPRIVATE:
		data[0] = dev->phy;
	case SIOCDEVPRIVATE + 1:
		read_mii_word(dev, dev->phy, (data[1] & 0x1f), &data[3]);
		break;
	case SIOCDEVPRIVATE + 2:
		if (!capable(CAP_NET_ADMIN)) {
			res = -EPERM;
			break;
		}
		write_mii_word(dev, dev->phy, (data[1] & 0x1f), data[2]);
		break;
	default:
		res = -EOPNOTSUPP;
	}
	up(&dev->sem);

	return res;
}

static void *rtl8150_probe(struct usb_device *udev, unsigned int ifnum,
			   const struct usb_device_id *id)
{
	rtl8150_t *dev;
	struct net_device *netdev;

	dbg("<<");
	udev->config[0].bConfigurationValue = 1;
	if (usb_set_configuration(udev, udev->config[0].bConfigurationValue)) {
		err("usb_set_configuration() failed");
		return NULL;
	}
	dev = kmalloc(sizeof(rtl8150_t), GFP_KERNEL);
	if (!dev) {
		err("Out of memory");
		goto exit;
	} else
		memset(dev, 0, sizeof(rtl8150_t));
	usb_inc_dev_use(udev);

	netdev = init_etherdev(NULL, 0);
	if (!netdev) {
		kfree(dev);
		err("Oh boy, out of memory again?!?");
		usb_dec_dev_use(udev);
		dev = NULL;
		goto exit;
	}

	init_MUTEX(&dev->sem);
	dev->udev = udev;
	dev->netdev = netdev;
	SET_MODULE_OWNER(netdev);
	netdev->priv = dev;
	netdev->open = rtl8150_open;
	netdev->stop = rtl8150_close;
	netdev->do_ioctl = rtl8150_ioctl;
	netdev->watchdog_timeo = RTL8150_TX_TIMEOUT;
	netdev->tx_timeout = rtl8150_tx_timeout;
	netdev->hard_start_xmit = rtl8150_start_xmit;
	netdev->set_multicast_list = rtl8150_set_multicast;
	netdev->get_stats = rtl8150_netdev_stats;
	netdev->mtu = RTL8150_MTU;
	dev->intr_interval = 100;	/* 100ms */

	if (rtl8150_reset(dev) || !alloc_all_urbs(dev)) {
		err("couldn't reset the device");
		free_all_urbs(dev);
		unregister_netdev(dev->netdev);
		kfree(netdev);
		kfree(dev);
		usb_dec_dev_use(udev);
		dev = NULL;
		goto exit;
	}

	netif_carrier_off(dev->netdev);
	set_ethernet_addr(dev);
	/* let's not be very nasty :-) */
	info("%s: rtl8150 is detected", netdev->name);
exit:
	dbg(">>");
	return dev;
}

static void rtl8150_disconnect(struct usb_device *udev, void *ptr)
{
	rtl8150_t *dev;

	dbg("<<");
	dev = ptr;
	set_bit(RTL8150_UNPLUG, &dev->flags);
	unregister_netdev(dev->netdev);
	free_all_urbs(dev);
	usb_dec_dev_use(udev);
	dev->netdev->priv = NULL;
	kfree(dev->netdev);
	dev->netdev = NULL;
	kfree(dev);
	dbg(">>");
}

static int __init usb_rtl8150_init(void)
{
	int res;
	dbg("<<");
	info(DRIVER_DESC " " DRIVER_VERSION);
	res = usb_register(&rtl8150_driver);
	dbg(">>");
	return res;
}

static void __exit usb_rtl8150_exit(void)
{
	dbg("<<");
	usb_deregister(&rtl8150_driver);
	dbg(">>");
}

module_init(usb_rtl8150_init);
module_exit(usb_rtl8150_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
